# this file is automatically generated and overwritten by CI process
# to modify the content, modify the CI workflow file - .github/workflows/main.yml, job: "Bump patch version"
__version__ = "0.1.11"
